from dataclasses import dataclass
from typing import Callable, Dict, List, Union

from PIL import Image
import cv2 as cv
import numpy as np
import skimage.segmentation as seg
import torch.nn as nn

import explainer.models as _models
from explainer.models.base import VisionModel

__all__ = [
    "list_segmentation_methods",
    "get_segmentation_method",
]


class SegmentationWrapper(VisionModel):
    def __init__(
        self,
        fn: Union[Callable[[np.ndarray], np.ndarray], nn.Module],
        default_kwargs: Dict = dict(),
    ):
        super().__init__()
        self.fn = fn  # the function or nn.Module to wrap
        self.default_kwargs = default_kwargs

    def forward(self, img: np.ndarray, maps: List[_models.XAI_Map], **kwargs):
        kwargs = {**self.default_kwargs, **kwargs}
        # TODO: handle segmentationmap. If maps contains a segmentationmap, return this segmentation and don't call self.fn
        while True:
            try:
                x = self.fn(img, **kwargs)
                if x.min() == 0:
                    x += 1  # keep 0 for background
                return x
            except TypeError as e:
                if "unexpected keyword argument" in str(e):
                    # Remove unexpected keyword argument
                    unexpected_kwarg = str(e).split("'")[1]
                    kwargs.pop(unexpected_kwarg)

                    print(f"Removed unexpected keyword argument {unexpected_kwarg}.")

                else:
                    raise e


@dataclass(frozen=True)
class RegisteredSegmentationMethod:
    handle: Callable
    name: str
    is_builder: bool  # Flag to indicate if the handle is a builder function or can be called directly
    default_kwargs: Dict
    builder_kwargs: Dict

    def __post_init__(self):
        if not self.is_builder:
            assert (
                len(self.builder_kwargs) == 0
            ), "builder_kwargs should only be specified if is_builder is True."

    def initialize(self):
        handle = self.handle(**self.builder_kwargs) if self.is_builder else self.handle

        return SegmentationWrapper(handle, self.default_kwargs)


REGISTERED_SEGMENTATION_METHODS = {}  # type: Dict[str, RegisteredSegmentationMethod]


def register_segmentation_method(
    name: str = None,
    is_builder: bool = False,
    default_kwargs: Dict = dict(),
    builder_kwargs: Dict = dict(),
) -> Callable[[Callable], Callable]:
    """
    Registers a segmentation method.

    Args:
        name: The name of the segmentation method. If None, the name of the function is used.
        is_builder: A flag to indicate if the handle is a builder function or can be called directly.
        default_kwargs: Default keyword arguments for the segmentation method.
        builder_kwargs: Keyword arguments for the builder function.

    Returns:
        A decorator that registers the segmentation method.
    """

    def decorator(fn: Callable):
        global REGISTERED_SEGMENTATION_METHODS
        nonlocal name

        if name is None:
            name = fn.__name__

        if name in REGISTERED_SEGMENTATION_METHODS:
            raise ValueError(
                f"A segmentation method with the name {name} has already been registered."
            )

        REGISTERED_SEGMENTATION_METHODS[name] = RegisteredSegmentationMethod(
            handle=fn,
            name=name,
            is_builder=is_builder,
            default_kwargs=default_kwargs,
            builder_kwargs=builder_kwargs,
        )
        return fn

    return decorator


def list_segmentation_methods():
    return list(REGISTERED_SEGMENTATION_METHODS.keys())


def get_segmentation_method(method_name: str) -> SegmentationWrapper:
    return REGISTERED_SEGMENTATION_METHODS[method_name].initialize()


register_segmentation_method(default_kwargs={"n_segments": 100},)(
    seg.slic
)  # allows mask

register_segmentation_method()(seg.quickshift)  # no mask

register_segmentation_method()(seg.watershed)  # allows mask


@register_segmentation_method()
def chan_vese(img: np.ndarray):
    img = cv.cvtColor(img, cv.COLOR_RGB2GRAY)
    return seg.chan_vese(img)  # no mask


class SAM_Wrapper(nn.Module):  # no mask
    def __init__(self, sam):
        super().__init__()
        self.sam = sam

    def forward(self, img: np.ndarray):
        img = Image.fromarray(img)

        return self.sam.predict(img)


for segmentation_model in _models.list_models(model_type="segmentation_model"):

    @register_segmentation_method(name=segmentation_model, is_builder=True)
    def build_sam():
        sam = _models.get_model(segmentation_model)
        sam = SAM_Wrapper(sam)
        return sam


def show_segmentation_on_image(
    img: np.ndarray, segmentation: np.ndarray
) -> Image.Image:
    """
    Shows the segmentation on the image.

    Args:
        img: The image.
        segmentation: The segmentation.

    Returns:
        The image with the segmentation overlayed.
    """
    from skimage.segmentation import mark_boundaries

    def _prepare_img(img) -> np.ndarray:
        if img.max() <= 1:
            img *= 255
            img = np.array(img, dtype=np.uint8)
        else:
            img = np.array(img, dtype=np.uint8)

        return img

    fig = mark_boundaries(img, segmentation)
    fig = _prepare_img(fig)

    return Image.fromarray(fig)


if __name__ == "__main__":
    print(list_segmentation_methods()[:-3])

    image = np.array(
        Image.open(
            r"C:\Users\Raphael\Desktop\Dev\xai-first-project\explainer\example_images\2007_000027.jpg"
        )
    )

    test_kwargs = {
        "FORCE_ERROR": None,  # This should be ignored
        "n_segments": 250,
        "compactness": 10,
        "max_iter": 10,
    }

    results = []

    for segmentation_method in list_segmentation_methods()[:4]:
        print(segmentation_method)
        segmentation_method = get_segmentation_method(segmentation_method)

        print(type(segmentation_method))

        result = segmentation_method(image, **test_kwargs)

        print(result.shape)
        print(result)

        results.append((segmentation_method, result))

    for segmentation_method, result in results:
        show_segmentation_on_image(image, result).show()
